@@ -17,7 +17,7 @@ from ..errors import EWSWarning, TransportError, SOAPError, ErrorTimeoutExpired,
     ErrorNoPublicFolderReplicaAvailable, MalformedResponseError, ErrorExceededConnectionCount, \
     SessionPoolMinSizeReached, ErrorIncorrectSchemaVersion, ErrorInvalidRequest, ErrorCorruptData, \
     ErrorCannotEmptyFolder, ErrorDeleteDistinguishedFolder, ErrorInvalidSubscription, ErrorInvalidWatermark, \
-    ErrorInvalidSyncStateData
+    ErrorInvalidSyncStateData, ErrorNameResolutionNoResults, ErrorNameResolutionMultipleResults
 from ..properties import FieldURI, IndexedFieldURI, ExtendedFieldURI, ExceptionFieldURI, ItemId
 from ..transport import wrap
 from ..util import chunkify, create_element, add_xml_child, get_xml_attr, to_xml, post_ratelimited, \
@@ -51,6 +51,8 @@ KNOWN_EXCEPTIONS = (
     ErrorItemNotFound,
     ErrorMailboxMoveInProgress,
     ErrorMailboxStoreUnavailable,
+    ErrorNameResolutionMultipleResults,
+    ErrorNameResolutionNoResults,
     ErrorNonExistentMailbox,
     ErrorNoPublicFolderReplicaAvailable,
     ErrorNoRespondingCASInDestinationSite,
@@ -80,8 +82,6 @@ class EWSService(metaclass=abc.ABCMeta):
     WARNINGS_TO_IGNORE_IN_RESPONSE = ()
     # The exception type to raise when all attempted API versions failed
     NO_VALID_SERVER_VERSIONS = ErrorInvalidServerVersion
-    # Controls whether the HTTP request should be streaming or fetch everything at once
-    streaming = False
     # Marks the version from which the service was introduced
     supported_from = None
     # Marks services that support paging of requested items
@@ -102,6 +102,8 @@ class EWSService(metaclass=abc.ABCMeta):
         self.protocol = protocol
         # Allow a service to override the default protocol timeout. Useful for streaming services
         self.timeout = timeout
+        # Controls whether the HTTP request should be streaming or fetch everything at once
+        self.streaming = False
         # Streaming connection variables
         self._streaming_session = None
         self._streaming_response = None
